import logging
from datetime import datetime
from threading import RLock
from typing import Any, Callable, Iterator, List, Optional, Protocol
from uuid import UUID, uuid4

from .schemas import (
    Span,
    SpanAttributes,
    SpanContext,
    SpanConversationAttributes,
    SpanEvent,
    SpanID,
    SpanKind,
    SpanStatusCode,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SpanExporter(Protocol):
    def export(self, span: Span) -> None:
        ...


class Tracer:
    """
    Tracer creates spans containing more information about what is happening for
    a given operation, such as a request in a service.

    OpenTelemetry Inspiration:
    https://opentelemetry.io/docs/concepts/signals/traces/#tracer
    """

    span_buffer: List[Span]
    on_append: Optional[Callable[[List[Span]], None]]

    def __init__(
        self,
        exporter: Optional[SpanExporter] = None,
        on_append: Optional[Callable[[List[Span]], None]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Create a new Tracer. A Tracer's main purpose is to create spans.
        Serialization should be handled by a separate component.

        Args:
            on_append:
                A callback function that will be called when a span is
                created and appended to the buffer. This is useful for
                serializing data to a file or sending it to a remote server.
        """
        self.span_buffer = []
        self.on_append = on_append
        self._exporter: Optional[SpanExporter] = exporter
        self._lock = RLock()
        super().__init__(*args, **kwargs)

    def create_span(
        self,
        name: str,
        span_kind: SpanKind,
        start_time: datetime,
        end_time: Optional[datetime] = None,
        status_code: SpanStatusCode = SpanStatusCode.UNSET,
        status_message: Optional[str] = "",
        parent_id: Optional[SpanID] = None,
        trace_id: Optional[UUID] = None,
        attributes: Optional[SpanAttributes] = None,
        events: Optional[List[SpanEvent]] = None,
        conversation: Optional[SpanConversationAttributes] = None,
        span_id: Optional[UUID] = None,
    ) -> Span:
        """
        create_span creates a new span with the given name and options.
        """
        # If no trace_id is provided, generate a new one
        if trace_id is None:
            trace_id = uuid4()

        # If no attributes are provided, create an empty dict
        if attributes is None:
            attributes = {}

        # If no events are provided, create an empty list
        if events is None:
            events = []

        span = Span(
            name=name,
            context=SpanContext(trace_id=trace_id, span_id=span_id or uuid4()),
            span_kind=span_kind,
            parent_id=parent_id,
            start_time=start_time,
            end_time=end_time,
            status_code=status_code,
            status_message=status_message if status_message is not None else "",
            attributes=attributes,
            events=events,
            conversation=conversation,
        )

        if self._exporter:
            self._exporter.export(span)

        with self._lock:
            self.span_buffer.append(span)

        if self.on_append is not None:
            self.on_append(self.span_buffer)
        return span

    def get_spans(self) -> Iterator[Span]:
        """
        Returns the spans stored in the tracer. This is useful if you are running
        in a notebook environment and you want to inspect the spans.
        """
        yield from self.span_buffer



import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from phoenix.experimental.evals.models.base import BaseEvalModel

if TYPE_CHECKING:
    from tiktoken import Encoding

logger = logging.getLogger(__name__)


@dataclass
class LiteLLMModel(BaseEvalModel):
    model_name: str = "gpt-3.5-turbo"
    """The model name to use."""
    temperature: float = 0.0
    """What sampling temperature to use."""
    max_tokens: int = 256
    """The maximum number of tokens to generate in the completion."""
    top_p: float = 1
    """Total probability mass of tokens to consider at each step."""
    num_retries: int = 6
    """Maximum number to retry a model if an RateLimitError, OpenAIError, or
    ServiceUnavailableError occurs."""
    request_timeout: int = 60
    """Maximum number of seconds to wait when retrying."""
    model_kwargs: Dict[str, Any] = field(default_factory=dict)
    """Model specific params"""

    # non-LiteLLM params
    retry_min_seconds: int = 10
    """Minimum number of seconds to wait when retrying."""
    max_content_size: Optional[int] = None
    """If you're using a fine-tuned model, set this to the maximum content size"""

    def __post_init__(self) -> None:
        self._init_environment()
        self._init_model_encoding()

    def _init_environment(self) -> None:
        try:
            import litellm
            from litellm import validate_environment

            self._litellm = litellm
            env_info = validate_environment(self._litellm.utils.get_llm_provider(self.model_name))

            if not env_info["keys_in_environment"]:
                raise RuntimeError(
                    f"Missing environment variable(s): '{str(env_info['missing_keys'])}', for "
                    f"model: {self.model_name}. \nFor additional information about the right "
                    "environment variables for specific model providers:\n"
                    "https://docs.litellm.ai/docs/completion/input#provider-specific-params."
                )
        except ImportError:
            self._raise_import_error(
                package_display_name="LiteLLM",
                package_name="litellm",
            )

    def _init_model_encoding(self) -> None:
        from litellm import decode, encode

        self._encoding = encode
        self._decoding = decode

    @property
    def max_context_size(self) -> int:
        context_size = self.max_content_size or self._litellm.get_max_tokens(self.model_name).get(
            "max_tokens", None
        )

        if context_size is None:
            raise ValueError(
                "Can't determine maximum context size. An unknown model name was "
                + f"used: {self.model_name}."
            )

        return context_size

    @property
    def encoder(self) -> "Encoding":
        raise NotImplementedError

    def get_tokens_from_text(self, text: str) -> List[int]:
        result: List[int] = self._encoding(model=self.model_name, text=text)
        return result

    def get_text_from_tokens(self, tokens: List[int]) -> str:
        return str(self._decoding(model=self.model_name, tokens=tokens))

    async def _async_generate(self, prompt: str, **kwargs: Dict[str, Any]) -> str:
        return self._generate(prompt, **kwargs)

    def _generate(self, prompt: str, **kwargs: Dict[str, Any]) -> str:
        messages = self._get_messages_from_prompt(prompt)
        return str(
            self._generate_with_retry(
                model=self.model_name,
                messages=messages,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                num_retries=self.num_retries,
                request_timeout=self.request_timeout,
                **self.model_kwargs,
            )
        )

    def _generate_with_retry(self, **kwargs: Any) -> Any:
        # Using default LiteLLM completion with retries = self.num_retries.

        response = self._litellm.completion(**kwargs)
        return response.choices[0].message.content

    def _get_messages_from_prompt(self, prompt: str) -> List[Dict[str, str]]:
        # LiteLLM requires prompts in the format of messages
        # messages=[{"content": "ABC?","role": "user"}]
        return [{"content": prompt, "role": "user"}]



import json
from abc import ABC
from dataclasses import dataclass, field
from itertools import product
from pathlib import Path
from types import MappingProxyType
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Set, Tuple, Type, Union
from uuid import UUID, uuid4

import pandas as pd
from pandas.api.types import is_integer_dtype, is_numeric_dtype, is_string_dtype
from pyarrow import Table, parquet

from phoenix.config import TRACE_DATASET_DIR
from phoenix.exceptions import PhoenixException

EVAL_NAME_COLUMN_PREFIX = "eval."


class InvalidParquetMetadataError(PhoenixException):
    pass


class NeedsNamedIndex(ABC):
    index_names: Mapping[Tuple[str, ...], Callable[[Any], bool]]
    all_valid_index_name_sorted_combos: Set[Tuple[str, ...]]

    @classmethod
    def preferred_names(cls) -> List[str]:
        return [choices[0] for choices in cls.index_names.keys()]

    @classmethod
    def aliases(cls) -> Mapping[str, str]:
        return {alias: choices[0] for choices in cls.index_names.keys() for alias in choices[1:]}

    @classmethod
    def unalias(cls, name: str) -> str:
        return cls.aliases().get(name, name)

    @classmethod
    def is_valid_index_names(cls, names: Sequence[str]) -> bool:
        return (
            len(names) == len(cls.index_names)
            and tuple(sorted(names)) in cls.all_valid_index_name_sorted_combos
        )

    @classmethod
    def find_valid_index_names(cls, dtypes: "pd.Series[Any]") -> Optional[List[str]]:
        valid_names = []
        for names, check_type in cls.index_names.items():
            for name in names:
                if name in dtypes.index and check_type(dtypes[name]):
                    valid_names.append(name)
                    break
            else:
                return None
        return valid_names


class NeedsResultColumns(ABC):
    result_column_names: Mapping[str, Callable[[Any], bool]] = MappingProxyType(
        {
            "score": is_numeric_dtype,
            "label": is_string_dtype,
            "explanation": is_string_dtype,
        }
    )

    @classmethod
    def is_valid_result_columns(cls, dtypes: "pd.Series[Any]") -> bool:
        names = cls.result_column_names.keys()
        intersection = dtypes.index.intersection(names)  # type: ignore
        if not len(intersection):
            return False
        for name in intersection:
            check_type = cls.result_column_names[name]
            if not check_type(dtypes[name]):
                return False
        return True


@dataclass(frozen=True)
class Evaluations(NeedsNamedIndex, NeedsResultColumns, ABC):
    eval_name: str  # The name for the evaluation, e.g. 'toxicity'
    dataframe: pd.DataFrame = field(repr=False)
    id: UUID = field(init=False, default_factory=uuid4)

    def __len__(self) -> int:
        return len(self.dataframe)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(eval_name={self.eval_name!r}, "
            f"dataframe=<rows: {len(self.dataframe)!r}>)"
        )

    def __dir__(self) -> List[str]:
        return ["get_dataframe"]

    def get_dataframe(self, prefix_columns_with_name: bool = True) -> pd.DataFrame:
        """
        Returns a copy of the dataframe with the evaluation annotations

        Parameters
        __________
        prefix_columns_with_name: bool
            if True, the columns will be prefixed with the eval_name, e.g. 'eval.toxicity.value'
        """
        if prefix_columns_with_name:
            prefix = f"{EVAL_NAME_COLUMN_PREFIX}{self.eval_name}."
            return self.dataframe.add_prefix(prefix)
        return self.dataframe.copy(deep=False)

    def __bool__(self) -> bool:
        return not self.dataframe.empty

    def __post_init__(self) -> None:
        dataframe = (
            pd.DataFrame() if self.dataframe.empty else self._clean_dataframe(self.dataframe)
        )
        object.__setattr__(self, "dataframe", dataframe)

    def _clean_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        # Ensure column names are strings.
        column_names = dataframe.columns.astype(str)
        dataframe = dataframe.set_axis(column_names, axis=1)

        # If the dataframe contains the index columns, set the index to those columns
        if not self.is_valid_index_names(dataframe.index.names) and (
            index_names := self.find_valid_index_names(dataframe.dtypes)
        ):
            dataframe = dataframe.set_index(index_names)

        # Validate that the dataframe is indexed correctly.
        if not self.is_valid_index_names(dataframe.index.names):
            raise ValueError(
                f"The dataframe index must be {self.preferred_names()} but was "
                f"'{dataframe.index.name or dataframe.index.names}'"
            )

        # Validate that the dataframe contains result columns of appropriate types.
        if not self.is_valid_result_columns(dataframe.dtypes):
            raise ValueError(
                f"The dataframe must contain one of these columns with appropriate "
                f"value types: {self.result_column_names.keys()} "
            )

        # Un-alias to the preferred names.
        preferred_names = [self.unalias(name) for name in dataframe.index.names]
        dataframe = dataframe.rename_axis(preferred_names, axis=0)

        # Drop the unnecessary columns.
        result_column_names = dataframe.columns.intersection(self.result_column_names.keys())  # type: ignore
        return dataframe.loc[:, result_column_names]  # type: ignore

    def __init_subclass__(
        cls,
        index_names: Mapping[Tuple[str, ...], Callable[[Any], bool]],
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        cls.index_names = index_names
        cls.all_valid_index_name_sorted_combos = set(
            tuple(sorted(prod)) for prod in product(*cls.index_names.keys())
        )

    def to_parquet(self, directory: Optional[Union[str, Path]] = None) -> Path:
        """Persists the evaluations to a parquet file.

        Args:
            directory (Optional[Union[str, Path]], optional): An optional path
            to a directory where the parquet file will be saved. If not
            provided, the parquet file will be saved to a default location.

        Returns:
            Path: The path to the parquet file, including a randomly generated
            filename.
        """
        directory = Path(directory) if directory else TRACE_DATASET_DIR
        path = directory / f"evaluations-{self.id}.parquet"
        table = Table.from_pandas(self.dataframe)
        table = table.replace_schema_metadata(
            {
                **(table.schema.metadata or {}),
                # explicitly encode keys and values, which are automatically encoded regardless
                b"arize": json.dumps(
                    {
                        "eval_id": str(self.id),
                        "eval_name": self.eval_name,
                        "eval_type": self.__class__.__name__,
                    }
                ).encode("utf-8"),
            }
        )
        parquet.write_table(table, path)
        return path

    @classmethod
    def from_parquet(cls, path: Union[str, Path]) -> "Evaluations":
        """Loads the evaluations from a parquet file.

        Args:
            path (Union[str, Path]): Path to a persisted evaluations parquet
            file.

        Returns:
            Evaluations: The loaded evaluations. The type of the returned
            evaluations will be the same as the type of the evaluations that
            were originally persisted.
        """
        schema = parquet.read_schema(path)
        eval_id, eval_name, evaluations_cls = _parse_schema_metadata(schema.metadata)
        table = parquet.read_table(path)
        dataframe = table.to_pandas()
        evaluations = evaluations_cls(eval_name=eval_name, dataframe=dataframe)
        object.__setattr__(evaluations, "id", eval_id)
        return evaluations


@dataclass(frozen=True)
class SpanEvaluations(
    Evaluations,
    index_names=MappingProxyType({("context.span_id", "span_id"): is_string_dtype}),
):
    """
    SpanEvaluations is a set of evaluation annotations for a set of spans.
    SpanEvaluations encompasses the evaluation annotations for a single evaluation task
    such as toxicity or hallucinations.

    SpanEvaluations can be appended to TraceDatasets so that the spans and
    evaluations can be joined and analyzed together.

    Parameters
    __________
    eval_name: str
        the name of the evaluation, e.g. 'toxicity'
    dataframe: pandas.DataFrame
        the pandas dataframe containing the evaluation annotations Each row
        represents the evaluations on a span.

    Example
    _______

    DataFrame of evaluations for toxicity may look like:

    | span_id | score              | label              | explanation        |
    |---------|--------------------|--------------------|--------------------|
    | span_1  | 1                  | toxic              | bad language       |
    | span_2  | 0                  | non-toxic          | violence           |
    | span_3  | 1                  | toxic              | discrimination     |
    """


@dataclass(frozen=True)
class DocumentEvaluations(
    Evaluations,
    index_names=MappingProxyType(
        {
            ("context.span_id", "span_id"): is_string_dtype,
            ("document_position", "position"): is_integer_dtype,
        }
    ),
):
    """
    DocumentEvaluations is a set of evaluation annotations for a set of documents.
    DocumentEvaluations encompasses the evaluation annotations for a single evaluation task
    such as relevance.

    Parameters
    __________
    eval_name: str
        the name of the evaluation, e.g. 'relevance'
    dataframe: pandas.DataFrame
        the pandas dataframe containing the evaluation annotations. Each row
        represents the evaluations on a document.

    Example
    _______

    DataFrame of document evaluations for relevance may look like:

    | span_id | position | score | label      | explanation  |
    |---------|----------|-------|------------|--------------|
    | span_1  | 0        | 1     | relevant   | it's apropos |
    | span_1  | 1        | 1     | relevant   | it's germane |
    | span_2  | 0        | 0     | irrelevant | it's rubbish |
    """

    def _clean_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        dataframe = super()._clean_dataframe(dataframe)
        if dataframe.index.names != self.preferred_names():
            return dataframe.swaplevel()
        return dataframe


@dataclass(frozen=True)
class TraceEvaluations(
    Evaluations,
    index_names=MappingProxyType({("context.trace_id", "trace_id"): is_string_dtype}),
):
    ...


def _parse_schema_metadata(metadata: Dict[bytes, Any]) -> Tuple[UUID, str, Type[Evaluations]]:
    """Validates and parses the schema metadata. Raises an exception if the
    metadata is invalid.

    Args:
        metadata (Dict[bytes, Any]): A dictionary of schema metadata from a
        parquet file.

    Returns:
        Tuple[str, ModuleType]: The evaluation name and the evaluations class.
    """
    if not (arize_metadata_json := metadata.get(b"arize")):
        raise InvalidParquetMetadataError('Schema metadata is missing "arize" key')
    try:
        arize_metadata = json.loads(arize_metadata_json)
    except json.JSONDecodeError as err:
        raise InvalidParquetMetadataError(
            'Encountered invalid JSON string under "arize" key'
        ) from err
    evaluations_classes = {subclass.__name__: subclass for subclass in Evaluations.__subclasses__()}
    if not (
        isinstance(arize_metadata, dict)
        and (eval_id := _to_uuid(arize_metadata.get("eval_id")))
        and isinstance(eval_name := arize_metadata.get("eval_name"), str)
        and (eval_type := arize_metadata.get("eval_type"))
        and (evaluations_cls := evaluations_classes.get(eval_type))
    ):
        raise InvalidParquetMetadataError(f"Invalid Arize metadata: {arize_metadata}")
    return eval_id, eval_name, evaluations_cls


def _to_uuid(value: Any) -> Optional[UUID]:
    """
    Converts an input to a UUID if possible, otherwise returns None.

    Args:
        value (Any): The value to convert to a UUID.

    Returns:
        Optional[UUID]: A UUID if the value could be converted, otherwise None.
    """
    try:
        return UUID(value)
    except Exception:
        return None



from datetime import datetime, timezone
from typing import TYPE_CHECKING, Generator, List
from uuid import UUID

from llama_index.callbacks.schema import TIMESTAMP_FORMAT
from llama_index.response.schema import StreamingResponse

from phoenix.trace.schemas import SpanKind, SpanStatusCode
from phoenix.trace.semantic_conventions import OUTPUT_VALUE
from phoenix.trace.tracer import Tracer

if TYPE_CHECKING:
    from phoenix.trace.llama_index.callback import CBEventData

_LOCAL_TZINFO = datetime.now().astimezone().tzinfo


class TokenGenInstrumentor:
    def __init__(
        self, stream: Generator[str, None, None], tracer: Tracer, event_data: "CBEventData"
    ):
        self._stream = stream
        self._token_stream: List[str] = []
        self._finished = False
        self._tracer = tracer
        self._event_data = event_data

    def __iter__(self) -> "TokenGenInstrumentor":
        return self

    def __next__(self) -> str:
        if self._finished:
            raise StopIteration

        try:
            value = next(self._stream)
            self._token_stream.append(value)
            return value
        except StopIteration:
            self._finished = True
            self._handle_end_of_stream()
            raise

    def _handle_end_of_stream(self) -> None:
        # Handle the end-of-stream logic here
        parent_id = self._event_data.parent_id
        output = "".join(self._token_stream)
        start_event = self._event_data.start_event
        if start_event:
            start_time = _timestamp_to_tz_aware_datetime(start_event.time)
        else:
            start_time = datetime.now(timezone.utc)
        attributes = self._event_data.attributes
        attributes.update({OUTPUT_VALUE: output})
        self._tracer.create_span(
            name=self._event_data.name if self._event_data.name else "",
            span_kind=SpanKind.LLM,
            trace_id=self._event_data.trace_id,
            start_time=start_time,
            end_time=datetime.now(timezone.utc),
            status_code=SpanStatusCode.OK,
            status_message="",
            parent_id=UUID(parent_id) if parent_id else None,
            attributes=self._event_data.attributes,
            events=[],
            conversation=None,
            span_id=UUID(self._event_data.span_id),
        )


def instrument_streaming_response(
    response: StreamingResponse,
    tracer: Tracer,
    event_data: "CBEventData",
) -> StreamingResponse:
    if response.response_gen is not None:
        response.response_gen = TokenGenInstrumentor(response.response_gen, tracer, event_data)  # type: ignore
    return response


def _timestamp_to_tz_aware_datetime(timestamp: str) -> datetime:
    """Converts a timestamp string to a timezone-aware datetime."""
    return _tz_naive_to_tz_aware_datetime(_timestamp_to_tz_naive_datetime(timestamp))


def _timestamp_to_tz_naive_datetime(timestamp: str) -> datetime:
    """Converts a timestamp string to a timezone-naive datetime."""
    return datetime.strptime(timestamp, TIMESTAMP_FORMAT)


def _tz_naive_to_tz_aware_datetime(timestamp: datetime) -> datetime:
    """Converts a timezone-naive datetime to a timezone-aware datetime."""
    return timestamp.replace(tzinfo=_LOCAL_TZINFO)



import gzip
from typing import Protocol

import opentelemetry.proto.trace.v1.trace_pb2 as otlp
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import Response

from phoenix.trace.otel import encode
from phoenix.trace.schemas import Span
from phoenix.trace.span_json_decoder import json_to_span


class SupportsPutSpan(Protocol):
    def put(self, span: otlp.Span) -> None:
        ...


class SpanHandler(HTTPEndpoint):
    queue: SupportsPutSpan

    async def post(self, request: Request) -> Response:
        try:
            content_type = request.headers.get("content-type")
            if content_type == "application/x-protobuf":
                body = await request.body()
                content_encoding = request.headers.get("content-encoding")
                if content_encoding == "gzip":
                    body = gzip.decompress(body)
                otlp_span = otlp.Span()
                otlp_span.ParseFromString(body)
            else:
                span = json_to_span(await request.json())
                assert isinstance(span, Span)
                otlp_span = encode(span)
        except Exception:
            return Response(status_code=422)
        self.queue.put(otlp_span)
        return Response()



import logging
import weakref
from collections import defaultdict
from datetime import datetime, timezone
from queue import SimpleQueue
from threading import RLock, Thread
from types import MethodType
from typing import DefaultDict, Dict, List, Optional, Set, Tuple

import numpy as np
from google.protobuf.json_format import MessageToDict
from typing_extensions import TypeAlias, assert_never

import phoenix.trace.v1 as pb
from phoenix.trace.schemas import SpanID, TraceID

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

END_OF_QUEUE = None  # sentinel value for queue termination

EvaluationName: TypeAlias = str
DocumentPosition: TypeAlias = int


class Evals:
    def __init__(self) -> None:
        self._queue: "SimpleQueue[Optional[pb.Evaluation]]" = SimpleQueue()
        weakref.finalize(self, self._queue.put, END_OF_QUEUE)
        self._lock = RLock()
        self._trace_evaluations_by_name: DefaultDict[
            EvaluationName, Dict[TraceID, pb.Evaluation]
        ] = defaultdict(dict)
        self._evaluations_by_trace_id: DefaultDict[
            TraceID, Dict[EvaluationName, pb.Evaluation]
        ] = defaultdict(dict)
        self._span_evaluations_by_name: DefaultDict[
            EvaluationName, Dict[SpanID, pb.Evaluation]
        ] = defaultdict(dict)
        self._evaluations_by_span_id: DefaultDict[
            SpanID, Dict[EvaluationName, pb.Evaluation]
        ] = defaultdict(dict)
        self._span_evaluation_labels: DefaultDict[EvaluationName, Set[str]] = defaultdict(set)
        self._document_evaluations_by_span_id: DefaultDict[
            SpanID, DefaultDict[EvaluationName, Dict[DocumentPosition, pb.Evaluation]]
        ] = defaultdict(lambda: defaultdict(dict))
        self._document_evaluations_by_name: DefaultDict[
            EvaluationName, DefaultDict[SpanID, Dict[DocumentPosition, pb.Evaluation]]
        ] = defaultdict(lambda: defaultdict(dict))
        self._last_updated_at: Optional[datetime] = None
        self._start_consumer()

    def put(self, evaluation: pb.Evaluation) -> None:
        self._queue.put(evaluation)

    def _start_consumer(self) -> None:
        Thread(
            target=MethodType(
                self.__class__._consume_evaluations,
                weakref.proxy(self),
            ),
            daemon=True,
        ).start()

    def _consume_evaluations(self) -> None:
        while (item := self._queue.get()) is not END_OF_QUEUE:
            with self._lock:
                self._process_evaluation(item)

    def _process_evaluation(self, evaluation: pb.Evaluation) -> None:
        subject_id = evaluation.subject_id
        name = evaluation.name
        subject_id_kind = subject_id.WhichOneof("kind")
        if subject_id_kind == "document_retrieval_id":
            document_retrieval_id = subject_id.document_retrieval_id
            span_id = SpanID(document_retrieval_id.span_id)
            document_position = document_retrieval_id.document_position
            self._document_evaluations_by_span_id[span_id][name][document_position] = evaluation
            self._document_evaluations_by_name[name][span_id][document_position] = evaluation
        elif subject_id_kind == "span_id":
            span_id = SpanID(subject_id.span_id)
            self._evaluations_by_span_id[span_id][name] = evaluation
            self._span_evaluations_by_name[name][span_id] = evaluation
            if evaluation.result.HasField("label"):
                label = evaluation.result.label.value
                self._span_evaluation_labels[name].add(label)
        elif subject_id_kind == "trace_id":
            trace_id = TraceID(subject_id.trace_id)
            self._evaluations_by_trace_id[trace_id][name] = evaluation
            self._trace_evaluations_by_name[name][trace_id] = evaluation
        elif subject_id_kind is None:
            logger.warning(
                f"discarding evaluation with missing subject_id: {MessageToDict(evaluation)}"
            )
        else:
            assert_never(subject_id_kind)
        self._last_updated_at = datetime.now(timezone.utc)

    @property
    def last_updated_at(self) -> Optional[datetime]:
        return self._last_updated_at

    def get_span_evaluation(self, span_id: SpanID, name: str) -> Optional[pb.Evaluation]:
        with self._lock:
            span_evaluations = self._evaluations_by_span_id.get(span_id)
            return span_evaluations.get(name) if span_evaluations else None

    def get_span_evaluation_names(self) -> List[EvaluationName]:
        with self._lock:
            return list(self._span_evaluations_by_name)

    def get_document_evaluation_names(
        self,
        span_id: Optional[SpanID] = None,
    ) -> List[EvaluationName]:
        with self._lock:
            if span_id is None:
                return list(self._document_evaluations_by_name)
            document_evaluations = self._document_evaluations_by_span_id.get(span_id)
            return list(document_evaluations) if document_evaluations else []

    def get_span_evaluation_labels(self, name: EvaluationName) -> Tuple[str, ...]:
        with self._lock:
            labels = self._span_evaluation_labels.get(name)
            return tuple(labels) if labels else ()

    def get_span_evaluation_span_ids(self, name: EvaluationName) -> Tuple[SpanID, ...]:
        with self._lock:
            span_evaluations = self._span_evaluations_by_name.get(name)
            return tuple(span_evaluations.keys()) if span_evaluations else ()

    def get_evaluations_by_span_id(self, span_id: SpanID) -> List[pb.Evaluation]:
        with self._lock:
            evaluations = self._evaluations_by_span_id.get(span_id)
            return list(evaluations.values()) if evaluations else []

    def get_document_evaluation_span_ids(self, name: EvaluationName) -> Tuple[SpanID, ...]:
        with self._lock:
            document_evaluations = self._document_evaluations_by_name.get(name)
            return tuple(document_evaluations.keys()) if document_evaluations else ()

    def get_document_evaluations_by_span_id(self, span_id: SpanID) -> List[pb.Evaluation]:
        all_evaluations: List[pb.Evaluation] = []
        with self._lock:
            document_evaluations = self._document_evaluations_by_span_id.get(span_id)
            if not document_evaluations:
                return all_evaluations
            for evaluations in document_evaluations.values():
                all_evaluations.extend(evaluations.values())
        return all_evaluations

    def get_document_evaluation_scores(
        self,
        span_id: SpanID,
        evaluation_name: str,
        num_documents: int,
    ) -> List[float]:
        # num_documents is needed as argument because the document position values
        # are not checked during ingestion: e.g. if there exists a position value
        # of one trillion, we would not want to create a result that large.
        scores: List[float] = [np.nan] * num_documents
        with self._lock:
            document_evaluations = self._document_evaluations_by_span_id.get(span_id)
            if not document_evaluations:
                return scores
            evaluations = document_evaluations.get(evaluation_name)
            if not evaluations:
                return scores
            for document_position, evaluation in evaluations.items():
                result = evaluation.result
                if result.HasField("score") and document_position < num_documents:
                    scores[document_position] = result.score.value
        return scores



import json
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from inspect import BoundArguments, signature
from types import TracebackType
from typing import (
    Any,
    AsyncIterator,
    Callable,
    ContextManager,
    Coroutine,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

import openai
from openai import AsyncStream, Stream
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from typing_extensions import ParamSpec
from wrapt import ObjectProxy

from phoenix.trace.schemas import (
    MimeType,
    SpanAttributes,
    SpanEvent,
    SpanException,
    SpanKind,
    SpanStatusCode,
)
from phoenix.trace.semantic_conventions import (
    INPUT_MIME_TYPE,
    INPUT_VALUE,
    LLM_FUNCTION_CALL,
    LLM_INPUT_MESSAGES,
    LLM_INVOCATION_PARAMETERS,
    LLM_OUTPUT_MESSAGES,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    MESSAGE_CONTENT,
    MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON,
    MESSAGE_FUNCTION_CALL_NAME,
    MESSAGE_NAME,
    MESSAGE_ROLE,
    MESSAGE_TOOL_CALLS,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    TOOL_CALL_FUNCTION_ARGUMENTS_JSON,
    TOOL_CALL_FUNCTION_NAME,
)
from phoenix.trace.utils import get_stacktrace

from ..tracer import Tracer

ParameterSpec = ParamSpec("ParameterSpec")
GenericType = TypeVar("GenericType")
AsyncCallable = Callable[ParameterSpec, Coroutine[Any, Any, GenericType]]
Parameters = Mapping[str, Any]
OpenInferenceMessage = Dict[str, str]

INSTRUMENTED_ATTRIBUTE_NAME = "is_instrumented_with_openinference_tracer"


class RequestType(Enum):
    CHAT_COMPLETION = "chat_completion"
    COMPLETION = "completion"
    EMBEDDING = "embedding"


class OpenAIInstrumentor:
    def __init__(self, tracer: Optional[Tracer] = None) -> None:
        """Instruments your OpenAI client to automatically create spans for each API call.

        Args:
            tracer (Optional[Tracer], optional): A tracer to record and handle spans. If not
            provided, the default tracer will be used.
        """
        self._tracer = tracer or Tracer()

    def instrument(self) -> None:
        """
        Instruments your OpenAI client.
        """
        if not hasattr(openai.OpenAI, INSTRUMENTED_ATTRIBUTE_NAME):
            openai.OpenAI.request = _wrapped_openai_sync_client_request_function(  # type: ignore
                openai.OpenAI.request, self._tracer
            )
            setattr(
                openai.OpenAI,
                INSTRUMENTED_ATTRIBUTE_NAME,
                True,
            )
        if not hasattr(openai.AsyncOpenAI, INSTRUMENTED_ATTRIBUTE_NAME):
            openai.AsyncOpenAI.request = _wrapped_openai_async_client_request_function(  # type: ignore
                openai.AsyncOpenAI.request, self._tracer
            )
            setattr(
                openai.AsyncOpenAI,
                INSTRUMENTED_ATTRIBUTE_NAME,
                True,
            )


class ChatCompletionContext(ContextManager["ChatCompletionContext"]):
    """
    A context manager for creating spans for chat completion requests. The
    context manager extracts attributes from the input parameters and response
    from the API and records any exceptions that are raised.
    """

    def __init__(self, bound_arguments: BoundArguments, tracer: Tracer) -> None:
        """
        Initializes the context manager.

        Args:
            bound_arguments (BoundArguments): The arguments to the request
            function from which parameter attributes are extracted.

            tracer (Tracer): The tracer to use to create spans.
        """
        self.tracer = tracer
        self.start_time: Optional[datetime] = None
        self.status_code = SpanStatusCode.UNSET
        self.status_message = ""
        self.events: List[SpanEvent] = []
        self.attributes: SpanAttributes = dict()
        parameters = _parameters(bound_arguments)
        self.num_choices = parameters.get("n", 1)
        self.chat_completion_chunks: List[ChatCompletionChunk] = []
        self.stream_complete = False
        self._process_parameters(parameters)
        self._span_created = False

    def __enter__(self) -> "ChatCompletionContext":
        self.start_time = datetime.now(tz=timezone.utc)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_value is None:
            return
        self.status_code = SpanStatusCode.ERROR
        status_message = str(exc_value)
        self.status_message = status_message
        self.events.append(
            SpanException(
                message=status_message,
                timestamp=datetime.now(tz=timezone.utc),
                exception_type=type(exc_value).__name__,
                exception_stacktrace=get_stacktrace(exc_value),
            )
        )
        self.create_span()

    def process_response(self, response: Any) -> Any:
        """
        Processes the response from the OpenAI chat completions API call to extract attributes.

        Args:
            response (ChatCompletion): The chat completion object.
        """
        if isinstance(response, ChatCompletion):
            self._process_chat_completion(response)
        elif isinstance(response, Stream):
            return StreamWrapper(stream=response, context=self)
        elif isinstance(response, AsyncStream):
            return AsyncStreamWrapper(stream=response, context=self)
        elif hasattr(response, "parse") and callable(
            response.parse
        ):  # handle raw response by converting them to chat completions
            self._process_chat_completion(response.parse())
        self.status_code = SpanStatusCode.OK
        self.create_span()
        return response

    def create_span(self) -> None:
        """
        Creates a span from the context if one has not already been created.
        """
        if self._span_created:
            return
        self.tracer.create_span(
            name="OpenAI Chat Completion",
            span_kind=SpanKind.LLM,
            start_time=cast(datetime, self.start_time),
            end_time=datetime.now(tz=timezone.utc),
            status_code=self.status_code,
            status_message=self.status_message,
            attributes=self.attributes,
            events=self.events,
        )
        self._span_created = True

    def _process_chat_completion(self, chat_completion: ChatCompletion) -> None:
        """
        Processes a chat completion response to extract and add fields and
        attributes to the context.

        Args:
            chat_completion (ChatCompletion): Response object from the chat
            completions API.
        """
        for (
            attribute_name,
            get_chat_completion_attribute_fn,
        ) in _CHAT_COMPLETION_ATTRIBUTE_FUNCTIONS.items():
            if (attribute_value := get_chat_completion_attribute_fn(chat_completion)) is not None:
                self.attributes[attribute_name] = attribute_value

    def _process_parameters(self, parameters: Parameters) -> None:
        """
        Processes the input parameters to the chat completions API to extract
        and add fields and attributes to the context.

        Args:
            parameters (Parameters): Input parameters.
        """
        for (
            attribute_name,
            get_parameter_attribute_fn,
        ) in _PARAMETER_ATTRIBUTE_FUNCTIONS.items():
            if (attribute_value := get_parameter_attribute_fn(parameters)) is not None:
                self.attributes[attribute_name] = attribute_value


class StreamWrapper(ObjectProxy):  # type: ignore
    """
    A wrapper for streams of chat completion chunks that records events and
    creates the span upon completion of the stream or upon an exception.
    """

    def __init__(self, stream: Stream[ChatCompletionChunk], context: ChatCompletionContext) -> None:
        """Initializes the stream wrapper.

        Args:
            stream (Stream[ChatCompletionChunk]): The stream to wrap.

            context (ChatCompletionContext): The context used to store span
            fields and attributes.
        """
        super().__init__(stream)
        self._self_context = ChatCompletionStreamEventContext(context)

    def __next__(self) -> ChatCompletionChunk:
        """
        A wrapped __next__ method that records span stream events and updates
        the span upon completion of the stream or upon exception.

        Returns:
            ChatCompletionChunk: The forwarded chat completion chunk.
        """
        with self._self_context as context:
            chat_completion_chunk = next(self.__wrapped__)
            context.process_chat_completion_chunk(chat_completion_chunk)
            return cast(ChatCompletionChunk, chat_completion_chunk)

    def __iter__(self) -> Iterator[ChatCompletionChunk]:
        """
        A __iter__ method that bypasses the wrapped class' __iter__ method so
        that __iter__ is automatically instrumented using __next__.
        """
        return self


class AsyncStreamWrapper(ObjectProxy):  # type: ignore
    """
    A wrapper for asynchronous streams of chat completion chunks that records
    events and creates the span upon completion of the stream or upon an
    exception.
    """

    def __init__(
        self, stream: AsyncStream[ChatCompletionChunk], context: ChatCompletionContext
    ) -> None:
        """Initializes the stream wrapper.

        Args:
            stream (AsyncStream[ChatCompletionChunk]): The stream to wrap.

            context (ChatCompletionContext): The context used to store span
            fields and attributes.
        """
        super().__init__(stream)
        self._self_context = ChatCompletionStreamEventContext(context)

    async def __anext__(self) -> ChatCompletionChunk:
        """
        A wrapped __anext__ method that records span stream events and updates
        the span upon completion of the stream or upon exception.

        Returns:
            ChatCompletionChunk: The forwarded chat completion chunk.
        """
        with self._self_context as context:
            chat_completion_chunk = await self.__wrapped__.__anext__()
            context.process_chat_completion_chunk(chat_completion_chunk)
            return cast(ChatCompletionChunk, chat_completion_chunk)

    def __aiter__(self) -> AsyncIterator[ChatCompletionChunk]:
        """
        An __aiter__ method that bypasses the wrapped class' __aiter__ method so
        that __aiter__ is automatically instrumented using __anext__.
        """
        return self


class ChatCompletionStreamEventContext(ContextManager["ChatCompletionStreamEventContext"]):
    """
    A context manager that surrounds stream events in a stream of chat
    completions and processes each chat completion chunk.
    """

    def __init__(self, chat_completion_context: ChatCompletionContext) -> None:
        """Initializes the context manager.

        Args:
            chat_completion_context (ChatCompletionContext): The chat completion
            context storing span fields and attributes.
        """
        self._context = chat_completion_context

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if isinstance(exc_value, StopIteration) or isinstance(exc_value, StopAsyncIteration):
            self._context.stream_complete = True
            self._context.status_code = SpanStatusCode.OK
        elif exc_value is not None:
            self._context.stream_complete = True
            self._context.status_code = SpanStatusCode.ERROR
            status_message = str(exc_value)
            self._context.status_message = status_message
            self._context.events.append(
                SpanException(
                    message=status_message,
                    timestamp=datetime.now(tz=timezone.utc),
                    exception_type=type(exc_value).__name__,
                    exception_stacktrace=get_stacktrace(exc_value),
                )
            )
        if not self._context.stream_complete:
            return
        self._context.attributes = {
            **self._context.attributes,
            LLM_OUTPUT_MESSAGES: _accumulate_messages(
                chunks=self._context.chat_completion_chunks,
                num_choices=self._context.num_choices,
            ),  # type: ignore
            OUTPUT_VALUE: json.dumps(
                [chunk.dict() for chunk in self._context.chat_completion_chunks]
            ),
            OUTPUT_MIME_TYPE: MimeType.JSON,  # type: ignore
        }
        self._context.create_span()

    def process_chat_completion_chunk(self, chat_completion_chunk: ChatCompletionChunk) -> None:
        """
        Processes a chat completion chunk and adds relevant information to the
        context.

        Args:
            chat_completion_chunk (ChatCompletionChunk): The chat completion
            chunk to be processed.
        """
        if not self._context.chat_completion_chunks:
            self._context.events.append(
                SpanEvent(
                    name="First Token Stream Event",
                    timestamp=datetime.now(tz=timezone.utc),
                    attributes={},
                )
            )
        self._context.chat_completion_chunks.append(chat_completion_chunk)


def _wrapped_openai_sync_client_request_function(
    request_fn: Callable[ParameterSpec, GenericType], tracer: Tracer
) -> Callable[ParameterSpec, GenericType]:
    """
    Wraps the synchronous OpenAI client's request method to create spans for
    each API call.

    Args:
        request_fn (Callable[ParameterSpec, GenericType]): The request method on
        the OpenAI client.

        tracer (Tracer): The tracer to use to create spans.

    Returns:
        Callable[ParameterSpec, GenericType]: The wrapped request method.
    """
    call_signature = signature(request_fn)

    def wrapped(*args: Any, **kwargs: Any) -> Any:
        bound_arguments = call_signature.bind(*args, **kwargs)
        if _request_type(bound_arguments) is not RequestType.CHAT_COMPLETION:
            return request_fn(*args, **kwargs)
        with ChatCompletionContext(bound_arguments, tracer) as context:
            response = request_fn(*args, **kwargs)
            return context.process_response(response)

    return wrapped


def _wrapped_openai_async_client_request_function(
    request_fn: AsyncCallable[ParameterSpec, GenericType], tracer: Tracer
) -> AsyncCallable[ParameterSpec, GenericType]:
    """
    Wraps the asynchronous AsyncOpenAI client's request method to create spans
    for each API call.

    Args:
        request_fn (AsyncCallable[ParameterSpec, GenericType]): The request
        method on the AsyncOpenAI client.

        tracer (Tracer): The tracer to use to create spans.

    Returns:
        AsyncCallable[ParameterSpec, GenericType]: The wrapped request method.
    """
    call_signature = signature(request_fn)

    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        bound_arguments = call_signature.bind(*args, **kwargs)
        if _request_type(bound_arguments) is not RequestType.CHAT_COMPLETION:
            return await request_fn(*args, **kwargs)
        with ChatCompletionContext(bound_arguments, tracer) as context:
            response = await request_fn(*args, **kwargs)
            return context.process_response(response)

    return wrapped


def _input_value(parameters: Parameters) -> str:
    return json.dumps(parameters)


def _input_mime_type(_: Any) -> MimeType:
    return MimeType.JSON


def _llm_input_messages(parameters: Parameters) -> Optional[List[OpenInferenceMessage]]:
    if not (messages := parameters.get("messages")):
        return None
    return [_to_openinference_message(message, expects_name=True) for message in messages]


def _llm_invocation_parameters(
    parameters: Parameters,
) -> str:
    return json.dumps(parameters)


def _output_value(chat_completion: ChatCompletion) -> str:
    return chat_completion.json()


def _output_mime_type(_: Any) -> MimeType:
    return MimeType.JSON


def _llm_output_messages(chat_completion: ChatCompletion) -> List[OpenInferenceMessage]:
    return [
        _to_openinference_message(choice.message.dict(), expects_name=False)
        for choice in chat_completion.choices
    ]


def _llm_token_count_prompt(chat_completion: ChatCompletion) -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.prompt_tokens
    return None


def _llm_token_count_completion(chat_completion: ChatCompletion) -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.completion_tokens
    return None


def _llm_token_count_total(chat_completion: ChatCompletion) -> Optional[int]:
    if completion_usage := chat_completion.usage:
        return completion_usage.total_tokens
    return None


def _llm_function_call(
    chat_completion: ChatCompletion,
) -> Optional[str]:
    choices = chat_completion.choices
    choice = choices[0]
    if choice.finish_reason == "function_call" and (function_call := choice.message.function_call):
        return function_call.json()
    return None


def _request_type(bound_arguments: BoundArguments) -> Optional[RequestType]:
    options = bound_arguments.arguments["options"]
    url = options.url
    """Get OpenAI request type from URL, or returns None if the request type cannot be recognized"""
    if "chat/completions" in url:
        return RequestType.CHAT_COMPLETION
    if "completions" in url:
        return RequestType.COMPLETION
    if "embeddings" in url:
        return RequestType.EMBEDDING
    return None


def _to_openinference_message(
    message: Mapping[str, Any], *, expects_name: bool
) -> OpenInferenceMessage:
    """Converts an OpenAI input or output message to an OpenInference message.

    Args:
        message (Dict[str, Any]): The OpenAI message to be parsed.

        expects_name (bool): Whether to parse the "name" key in the OpenAI message. This key is
        sometimes included in "function"-role input messages to specify the function name, but is
        not included in output messages.

    Returns:
        OpenInferenceMessage: A message in OpenInference format.
    """
    openinference_message = {}
    if role := message.get("role"):
        openinference_message[MESSAGE_ROLE] = role
    if content := message.get("content"):
        openinference_message[MESSAGE_CONTENT] = content
    if function_call_data := message.get("function_call"):
        if function_name := function_call_data.get("name"):
            openinference_message[MESSAGE_FUNCTION_CALL_NAME] = function_name
        if function_arguments := function_call_data.get("arguments"):
            openinference_message[MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON] = function_arguments
    if tool_calls_data := message.get("tool_calls"):
        message_tool_calls = []
        for tool_call_data in tool_calls_data:
            if message_tool_call := dict(_get_tool_call(tool_call_data)):
                message_tool_calls.append(message_tool_call)
        if message_tool_calls:
            openinference_message[MESSAGE_TOOL_CALLS] = message_tool_calls
    if expects_name and (name := message.get("name")):
        openinference_message[MESSAGE_NAME] = name
    return openinference_message


def _get_tool_call(tool_call: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
    if function := tool_call.get("function"):
        if name := function.get("name"):
            yield TOOL_CALL_FUNCTION_NAME, name
        if arguments := function.get("arguments"):
            yield TOOL_CALL_FUNCTION_ARGUMENTS_JSON, arguments


_PARAMETER_ATTRIBUTE_FUNCTIONS: Dict[str, Callable[[Parameters], Any]] = {
    INPUT_VALUE: _input_value,
    INPUT_MIME_TYPE: _input_mime_type,
    LLM_INPUT_MESSAGES: _llm_input_messages,
    LLM_INVOCATION_PARAMETERS: _llm_invocation_parameters,
}
_CHAT_COMPLETION_ATTRIBUTE_FUNCTIONS: Dict[str, Callable[[ChatCompletion], Any]] = {
    OUTPUT_VALUE: _output_value,
    OUTPUT_MIME_TYPE: _output_mime_type,
    LLM_OUTPUT_MESSAGES: _llm_output_messages,
    LLM_TOKEN_COUNT_PROMPT: _llm_token_count_prompt,
    LLM_TOKEN_COUNT_COMPLETION: _llm_token_count_completion,
    LLM_TOKEN_COUNT_TOTAL: _llm_token_count_total,
    LLM_FUNCTION_CALL: _llm_function_call,
}


def _parameters(bound_arguments: BoundArguments) -> Parameters:
    """
    The parameters for the LLM call, e.g., temperature.

    Args:
        bound_arguments (BoundArguments): The bound arguments to the request function.

    Returns:
        Parameters: The parameters to the request function.
    """
    return cast(Parameters, bound_arguments.arguments["options"].json_data)


@dataclass
class StreamingFunctionCallData:
    """
    Stores function call data from a streaming chat completion.
    """

    name: Optional[str] = None
    argument_tokens: List[str] = field(default_factory=list)


def _accumulate_messages(
    chunks: List[ChatCompletionChunk], num_choices: int
) -> List[OpenInferenceMessage]:
    """
    Converts a list of chat completion chunks to a list of OpenInference messages.

    Args:
        chunks (List[ChatCompletionChunk]): The input chunks to be converted.

        num_choices (int): The number of choices in the chat completion (i.e.,
        the parameter `n` in the input parameters).

    Returns:
        List[OpenInferenceMessage]: The list of OpenInference messages.
    """
    if not chunks:
        return []
    content_token_lists: DefaultDict[int, List[str]] = defaultdict(list)
    function_calls: DefaultDict[int, StreamingFunctionCallData] = defaultdict(
        StreamingFunctionCallData
    )
    tool_calls: DefaultDict[int, DefaultDict[int, StreamingFunctionCallData]] = defaultdict(
        lambda: defaultdict(StreamingFunctionCallData)
    )
    roles: Dict[int, str] = {}
    for chunk in chunks:
        for choice in chunk.choices:
            choice_index = choice.index
            if content_token := choice.delta.content:
                content_token_lists[choice_index].append(content_token)
            if function_call := choice.delta.function_call:
                if function_name := function_call.name:
                    function_calls[choice_index].name = function_name
                if (function_argument_token := function_call.arguments) is not None:
                    function_calls[choice_index].argument_tokens.append(function_argument_token)
            if role := choice.delta.role:
                roles[choice_index] = role
            if choice.delta.tool_calls:
                for tool_call in choice.delta.tool_calls:
                    tool_index = tool_call.index
                    if not tool_call.function:
                        continue
                    if (name := tool_call.function.name) is not None:
                        tool_calls[choice_index][tool_index].name = name
                    if (arguments := tool_call.function.arguments) is not None:
                        tool_calls[choice_index][tool_index].argument_tokens.append(arguments)

    messages: List[OpenInferenceMessage] = []
    for choice_index in range(num_choices):
        message: Dict[str, Any] = {}
        if (role_ := roles.get(choice_index)) is not None:
            message[MESSAGE_ROLE] = role_
        if content_tokens := content_token_lists[choice_index]:
            message[MESSAGE_CONTENT] = "".join(content_tokens)
        if function_call_ := function_calls.get(choice_index):
            if (name := function_call_.name) is not None:
                message[MESSAGE_FUNCTION_CALL_NAME] = name
            if argument_tokens := function_call_.argument_tokens:
                message[MESSAGE_FUNCTION_CALL_ARGUMENTS_JSON] = "".join(argument_tokens)
        if tool_calls_ := tool_calls.get(choice_index):
            num_tool_calls = max(tool_index for tool_index in tool_calls_.keys()) + 1
            message[MESSAGE_TOOL_CALLS] = [{} for _ in range(num_tool_calls)]
            for tool_index, tool_call_ in tool_calls_.items():
                if (name := tool_call_.name) is not None:
                    message[MESSAGE_TOOL_CALLS][tool_index][TOOL_CALL_FUNCTION_NAME] = name
                if argument_tokens := tool_call_.argument_tokens:
                    message[MESSAGE_TOOL_CALLS][tool_index][
                        TOOL_CALL_FUNCTION_ARGUMENTS_JSON
                    ] = "".join(argument_tokens)
        messages.append(message)

    return messages



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

from phoenix.core.evals import Evals
from phoenix.core.model_schema import Model
from phoenix.core.traces import Traces


@dataclass
class Context:
    request: Union[Request, WebSocket]
    response: Optional[Response]
    model: Model
    export_path: Path
    corpus: Optional[Model] = None
    traces: Optional[Traces] = None
    evals: Optional[Evals] = None



import os
import tempfile
from pathlib import Path
from typing import List, Optional

# Phoenix environment variables
ENV_PHOENIX_PORT = "PHOENIX_PORT"
ENV_PHOENIX_HOST = "PHOENIX_HOST"
ENV_NOTEBOOK_ENV = "PHOENIX_NOTEBOOK_ENV"
ENV_PHOENIX_COLLECTOR_ENDPOINT = "PHOENIX_COLLECTOR_ENDPOINT"
"""
The endpoint traces and evals are sent to. This must be set if the Phoenix
server is running on a remote instance.
"""


def _get_temp_path() -> Path:
    """Get path to  directory in which to store temp phoenix server files."""
    return Path(tempfile.gettempdir()) / ".arize-phoenix"


def get_pids_path() -> Path:
    """Get path to directory in which to store temp phoenix instance pid files.
    This directory is used to track any currently running instances of Arize Phoenix
    on the host machine. The directory will be created if it does not exist.
    """
    path = _get_temp_path() / "pids"
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_running_pid() -> Optional[int]:
    for file in get_pids_path().iterdir():
        if file.name.isnumeric():
            return int(file.name)
    return None


for path in (
    ROOT_DIR := Path.home().resolve() / ".phoenix",
    EXPORT_DIR := ROOT_DIR / "exports",
    DATASET_DIR := ROOT_DIR / "datasets",
    TRACE_DATASET_DIR := ROOT_DIR / "trace_datasets",
):
    path.mkdir(parents=True, exist_ok=True)

PHOENIX_DIR = Path(__file__).resolve().parent
# Server config
SERVER_DIR = PHOENIX_DIR / "server"
# The host the server will run on after launch_app is called
HOST = "0.0.0.0"
# The port the server will run on after launch_app is called
PORT = 6006
# The prefix of datasets that are auto-assigned a name
GENERATED_DATASET_NAME_PREFIX = "phoenix_dataset_"


def get_exported_files(directory: Path) -> List[Path]:
    """
    Yields the list of paths of exported files.

    Parameters
    ----------
    directory: Path
        Disk location to search exported files.

    Returns
    -------
    list: List[Path]
        List of paths of the exported files.
    """
    return list(directory.glob("*.parquet"))


def get_env_port() -> int:
    return (
        int(port)
        if isinstance(port := os.getenv(ENV_PHOENIX_PORT), str) and port.isnumeric()
        else PORT
    )


def get_env_host() -> str:
    return os.getenv(ENV_PHOENIX_HOST) or HOST


def get_env_collector_endpoint() -> Optional[str]:
    return os.getenv(ENV_PHOENIX_COLLECTOR_ENDPOINT)



from typing import List, Optional, Protocol, Union, cast

import pandas as pd

from phoenix.trace.dsl import SpanQuery
from phoenix.trace.semantic_conventions import (
    DOCUMENT_CONTENT,
    DOCUMENT_SCORE,
    INPUT_VALUE,
    OUTPUT_VALUE,
    RETRIEVAL_DOCUMENTS,
)

INPUT = {"input": INPUT_VALUE}
OUTPUT = {"output": OUTPUT_VALUE}
IO = {**INPUT, **OUTPUT}

IS_ROOT = "parent_id is None"
IS_LLM = "span_kind == 'LLM'"
IS_RETRIEVER = "span_kind == 'RETRIEVER'"


class Session(Protocol):
    def query_spans(self, *query: SpanQuery) -> Optional[Union[pd.DataFrame, List[pd.DataFrame]]]:
        ...


def get_retrieved_documents(session: Session) -> pd.DataFrame:
    return cast(
        pd.DataFrame,
        session.query_spans(
            SpanQuery()
            .where(IS_RETRIEVER)
            .select("trace_id", **INPUT)
            .explode(
                RETRIEVAL_DOCUMENTS,
                reference=DOCUMENT_CONTENT,
                document_score=DOCUMENT_SCORE,
            )
        ),
    )


def get_qa_with_reference(session: Session) -> pd.DataFrame:
    return pd.concat(
        cast(
            List[pd.DataFrame],
            session.query_spans(
                SpanQuery().select(**IO).where(IS_ROOT),
                SpanQuery()
                .where(IS_RETRIEVER)
                .select(span_id="parent_id")
                .concat(
                    RETRIEVAL_DOCUMENTS,
                    reference=DOCUMENT_CONTENT,
                ),
            ),
        ),
        axis=1,
        join="inner",
    )



from enum import Enum
from functools import partial
from typing import Any, Iterable, Iterator, Optional, Protocol

import pandas as pd
import strawberry
from strawberry import UNSET
from typing_extensions import assert_never

import phoenix.trace.v1 as pb
from phoenix.core.traces import (
    END_TIME,
    START_TIME,
    ComputedAttributes,
)
from phoenix.server.api.types.SortDir import SortDir
from phoenix.trace import semantic_conventions
from phoenix.trace.schemas import Span, SpanID


@strawberry.enum
class SpanColumn(Enum):
    startTime = START_TIME
    endTime = END_TIME
    latencyMs = ComputedAttributes.LATENCY_MS.value
    tokenCountTotal = semantic_conventions.LLM_TOKEN_COUNT_TOTAL
    tokenCountPrompt = semantic_conventions.LLM_TOKEN_COUNT_PROMPT
    tokenCountCompletion = semantic_conventions.LLM_TOKEN_COUNT_COMPLETION
    cumulativeTokenCountTotal = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_TOTAL.value
    cumulativeTokenCountPrompt = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_PROMPT.value
    cumulativeTokenCountCompletion = ComputedAttributes.CUMULATIVE_LLM_TOKEN_COUNT_COMPLETION.value


@strawberry.enum
class EvalAttr(Enum):
    score = "score"
    label = "label"


@strawberry.input
class EvalResultKey:
    name: str
    attr: EvalAttr


class SupportsGetSpanEvaluation(Protocol):
    def get_span_evaluation(self, span_id: SpanID, name: str) -> Optional[pb.Evaluation]:
        ...


@strawberry.input(
    description="The sort key and direction for span connections. Must "
    "specify one and only one of either `col` or `evalResultKey`."
)
class SpanSort:
    col: Optional[SpanColumn] = UNSET
    eval_result_key: Optional[EvalResultKey] = UNSET
    dir: SortDir

    def __call__(
        self,
        spans: Iterable[Span],
        evals: Optional[SupportsGetSpanEvaluation] = None,
    ) -> Iterator[Span]:
        """
        Sorts the spans by the given key (column or eval) and direction
        """
        if self.eval_result_key:
            get_sort_key_value = partial(
                _get_eval_result_value,
                eval_name=self.eval_result_key.name,
                eval_attr=self.eval_result_key.attr,
                evals=evals,
            )
        else:
            get_sort_key_value = partial(
                _get_column_value,
                span_column=self.col or SpanColumn.startTime,
            )
        yield from pd.Series(spans, dtype=object).sort_values(
            key=lambda series: series.apply(get_sort_key_value),
            ascending=self.dir.value == SortDir.asc.value,
        )


def _get_column_value(span: Span, span_column: SpanColumn) -> Any:
    if span_column is SpanColumn.startTime:
        return span.start_time
    if span_column is SpanColumn.endTime:
        return span.end_time
    return span.attributes.get(span_column.value)


def _get_eval_result_value(
    span: Span,
    eval_name: str,
    eval_attr: EvalAttr,
    evals: Optional[SupportsGetSpanEvaluation] = None,
) -> Any:
    """
    Returns the evaluation result for the given span
    """
    if evals is None:
        return None
    span_id = span.context.span_id
    evaluation = evals.get_span_evaluation(span_id, eval_name)
    if evaluation is None:
        return None
    result = evaluation.result
    if eval_attr is EvalAttr.score:
        return result.score.value if result.HasField("score") else None
    if eval_attr is EvalAttr.label:
        return result.label.value if result.HasField("label") else None
    assert_never(eval_attr)



import logging
from abc import ABC, abstractmethod, abstractproperty
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Generator, List, Optional, Sequence, Type

from phoenix.experimental.evals.models.rate_limiters import RateLimiter

if TYPE_CHECKING:
    from tiktoken import Encoding


from tenacity import (
    RetryCallState,
    retry,
    retry_base,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)
from tqdm.asyncio import tqdm_asyncio
from tqdm.auto import tqdm
from typing_extensions import TypeVar

from phoenix.experimental.evals.utils.threads import to_thread
from phoenix.utilities.logging import printif

T = TypeVar("T", bound=type)


def is_list_of(lst: Sequence[object], tp: T) -> bool:
    return isinstance(lst, list) and all(isinstance(x, tp) for x in lst)


logger = logging.getLogger(__name__)

TQDM_BAR_FORMAT = (
    "Eta:{eta} |{bar}| {percentage:3.1f}% "
    "({n_fmt}/{total_fmt}) "
    "[{elapsed}<{remaining}, {rate_fmt}{postfix}]"
)


@contextmanager
def set_verbosity(
    model: "BaseEvalModel", verbose: bool = False
) -> Generator["BaseEvalModel", None, None]:
    try:
        _model_verbose_setting = model._verbose
        _rate_limiter_verbose_setting = model._rate_limiter._verbose
        model._verbose = verbose
        model._rate_limiter._verbose = verbose
        yield model
    finally:
        model._verbose = _model_verbose_setting
        model._rate_limiter._verbose = _rate_limiter_verbose_setting


@dataclass
class BaseEvalModel(ABC):
    _verbose: bool = False
    _rate_limiter: RateLimiter = field(default_factory=RateLimiter)

    def reload_client(self) -> None:
        pass

    def _retry(
        self,
        error_types: List[Type[BaseException]],
        min_seconds: int,
        max_seconds: int,
        max_retries: int,
    ) -> Callable[[Any], Any]:
        """Create a retry decorator for a given LLM and provided list of error types."""

        def log_retry(retry_state: RetryCallState) -> None:
            if fut := retry_state.outcome:
                exc = fut.exception()
            else:
                exc = None

            if exc:
                printif(
                    True,
                    f"Failed attempt {retry_state.attempt_number}: raised {repr(exc)}",
                )
            else:
                printif(True, f"Failed attempt {retry_state.attempt_number}")
            return None

        if not error_types:
            # default to retrying on all exceptions
            error_types = [Exception]

        retry_instance: retry_base = retry_if_exception_type(error_types[0])
        for error in error_types[1:]:
            retry_instance = retry_instance | retry_if_exception_type(error)
        return retry(
            reraise=True,
            stop=stop_after_attempt(max_retries),
            wait=wait_random_exponential(multiplier=1, min=min_seconds, max=max_seconds),
            retry=retry_instance,
            before_sleep=log_retry,
        )

    def __call__(self, prompt: str, instruction: Optional[str] = None, **kwargs: Any) -> str:
        """Run the LLM on the given prompt."""
        if not isinstance(prompt, str):
            raise TypeError(
                "Invalid type for argument `prompt`. Expected a string but found "
                f"{type(prompt)}. If you want to run the LLM on multiple prompts, use "
                "`generate` instead."
            )
        if instruction is not None and not isinstance(instruction, str):
            raise TypeError(
                "Invalid type for argument `instruction`. Expected a string but found "
                f"{type(instruction)}."
            )
        return self._generate(prompt=prompt, instruction=instruction, **kwargs)

    async def async_call(self, prompt: str, instruction: Optional[str] = None) -> str:
        """Run the LLM on the given prompt."""
        if not isinstance(prompt, str):
            raise TypeError(
                "Invalid type for argument `prompt`. Expected a string but found "
                f"{type(prompt)}. If you want to run the LLM on multiple prompts, use "
                "`generate` instead."
            )
        if instruction is not None and not isinstance(instruction, str):
            raise TypeError(
                "Invalid type for argument `instruction`. Expected a string but found "
                f"{type(instruction)}."
            )
        response = await self.agenerate(prompts=[prompt], instruction=instruction)
        return response[0]

    def generate(
        self, prompts: List[str], instruction: Optional[str] = None, **kwargs: Any
    ) -> List[str]:
        printif(self._verbose, f"Generating responses for {len(prompts)} prompts...")
        if extra_info := self.verbose_generation_info():
            printif(self._verbose, extra_info)
        if not is_list_of(prompts, str):
            raise TypeError(
                "Invalid type for argument `prompts`. Expected a list of strings "
                f"but found {type(prompts)}."
            )
        try:
            outputs = []
            for prompt in tqdm(prompts, bar_format=TQDM_BAR_FORMAT):
                output = self._generate(prompt=prompt, instruction=instruction, **kwargs)
                logger.info(f"Prompt: {prompt}\nInstruction: {instruction}\nOutput: {output}")
                outputs.append(output)

        except (KeyboardInterrupt, Exception) as e:
            raise e
        return outputs

    async def agenerate(self, prompts: List[str], instruction: Optional[str] = None) -> List[str]:
        if not is_list_of(prompts, str):
            raise TypeError(
                "Invalid type for argument `prompts`. Expected a list of strings "
                f"but found {type(prompts)}."
            )
        try:
            result: List[str] = await tqdm_asyncio.gather(
                *[self._agenerate(prompt=prompt, instruction=instruction) for prompt in prompts],
                bar_format=TQDM_BAR_FORMAT,
                ncols=100,
            )
        except (KeyboardInterrupt, Exception) as e:
            raise e
        return result

    def verbose_generation_info(self) -> str:
        # if defined, returns additional model-specific information to display if `generate` is
        # run with `verbose=True`
        return ""

    @abstractmethod
    async def _async_generate(self, prompt: str, **kwargs: Any) -> str:
        raise NotImplementedError

    @abstractmethod
    def _generate(self, prompt: str, **kwargs: Any) -> str:
        raise NotImplementedError

    async def _agenerate(self, prompt: str, instruction: Optional[str]) -> str:
        return str(await to_thread(self._generate, prompt=prompt, instruction=instruction))

    @staticmethod
    def _raise_import_error(
        package_name: str, package_display_name: str = "", package_min_version: str = ""
    ) -> None:
        if not package_display_name:
            package_display_name = package_name
        msg = (
            f"Could not import necessary dependencies to use {package_display_name}. "
            "Please install them with "
        )
        if package_min_version:
            msg += f"`pip install {package_name}>={package_min_version}`."
        else:
            msg += f"`pip install {package_name}`."
        raise ImportError(msg)

    @abstractmethod
    def get_tokens_from_text(self, text: str) -> List[int]:
        ...

    @abstractmethod
    def get_text_from_tokens(self, tokens: List[int]) -> str:
        ...

    @abstractproperty
    def max_context_size(self) -> int:
        ...

    @abstractproperty
    def encoder(self) -> "Encoding":
        ...



"""
Callback handler for emitting trace data in OpenInference tracing format.
OpenInference tracing is an open standard for capturing and storing
LLM Application execution logs.

It enables production LLMapp servers to seamlessly integrate with LLM
observability solutions such as Arize and Phoenix.

For more information on the specification, see
https://github.com/Arize-ai/open-inference-spec
"""
import json
import logging
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
    cast,
)
from uuid import UUID, uuid4

import llama_index
from llama_index.callbacks.base_handler import BaseCallbackHandler
from llama_index.callbacks.schema import (
    TIMESTAMP_FORMAT,
    CBEvent,
    CBEventType,
    EventPayload,
)
from llama_index.llms.types import ChatMessage, ChatResponse
from llama_index.response.schema import Response, StreamingResponse
from llama_index.tools import ToolMetadata
from typing_extensions import TypeGuard

from phoenix.trace.exporter import HttpExporter
from phoenix.trace.llama_index.streaming import (
    instrument_streaming_response as _instrument_streaming_response,
)
from phoenix.trace.schemas import (
    MimeType,
    Span,
    SpanEvent,
    SpanException,
    SpanID,
    SpanKind,
    SpanStatusCode,
    TraceID,
)
from phoenix.trace.semantic_conventions import (
    DOCUMENT_CONTENT,
    DOCUMENT_ID,
    DOCUMENT_METADATA,
    DOCUMENT_SCORE,
    EMBEDDING_EMBEDDINGS,
    EMBEDDING_MODEL_NAME,
    EMBEDDING_TEXT,
    EMBEDDING_VECTOR,
    INPUT_MIME_TYPE,
    INPUT_VALUE,
    LLM_INPUT_MESSAGES,
    LLM_INVOCATION_PARAMETERS,
    LLM_MODEL_NAME,
    LLM_OUTPUT_MESSAGES,
    LLM_PROMPT_TEMPLATE,
    LLM_PROMPT_TEMPLATE_VARIABLES,
    LLM_PROMPTS,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    MESSAGE_CONTENT,
    MESSAGE_NAME,
    MESSAGE_ROLE,
    MESSAGE_TOOL_CALLS,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    RERANKER_INPUT_DOCUMENTS,
    RERANKER_MODEL_NAME,
    RERANKER_OUTPUT_DOCUMENTS,
    RERANKER_QUERY,
    RERANKER_TOP_K,
    RETRIEVAL_DOCUMENTS,
    TOOL_CALL_FUNCTION_ARGUMENTS_JSON,
    TOOL_CALL_FUNCTION_NAME,
    TOOL_DESCRIPTION,
    TOOL_NAME,
    TOOL_PARAMETERS,
)
from phoenix.trace.tracer import SpanExporter, Tracer
from phoenix.trace.utils import extract_version_triplet, get_stacktrace
from phoenix.utilities.error_handling import graceful_fallback

LLAMA_INDEX_MINIMUM_VERSION_TRIPLET = (0, 9, 8)
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

CBEventID = str
_LOCAL_TZINFO = datetime.now().astimezone().tzinfo


@dataclass
class CBEventData:
    name: Optional[str] = field(default=None)
    event_type: Optional[CBEventType] = field(default=None)
    start_event: Optional[CBEvent] = field(default=None)
    end_event: Optional[CBEvent] = field(default=None)
    attributes: Dict[str, Any] = field(default_factory=dict)
    span_id: Optional[CBEventID] = field(default=None)
    parent_id: Optional[CBEventID] = field(default=None)
    trace_id: Optional[TraceID] = field(default=None)
    streaming_event: bool = field(default=False)

    def set_if_unset(self, key: str, value: Any) -> None:
        if not getattr(self, key):
            setattr(self, key, value)


ChildEventIds = Dict[CBEventID, List[CBEventID]]
EventData = Dict[CBEventID, CBEventData]


def payload_to_semantic_attributes(
    event_type: CBEventType,
    payload: Dict[str, Any],
    is_event_end: bool = False,
) -> Dict[str, Any]:
    """
    Converts a LLMapp payload to a dictionary of semantic conventions compliant attributes.
    """
    attributes: Dict[str, Any] = {}
    if event_type in (CBEventType.NODE_PARSING, CBEventType.CHUNKING):
        # TODO(maybe): handle these events
        return attributes
    if EventPayload.CHUNKS in payload and EventPayload.EMBEDDINGS in payload:
        attributes[EMBEDDING_EMBEDDINGS] = [
            {EMBEDDING_TEXT: text, EMBEDDING_VECTOR: vector}
            for text, vector in zip(payload[EventPayload.CHUNKS], payload[EventPayload.EMBEDDINGS])
        ]
    if event_type is not CBEventType.RERANKING and EventPayload.QUERY_STR in payload:
        attributes[INPUT_VALUE] = payload[EventPayload.QUERY_STR]
        attributes[INPUT_MIME_TYPE] = MimeType.TEXT
    if event_type is not CBEventType.RERANKING and EventPayload.NODES in payload:
        attributes[RETRIEVAL_DOCUMENTS] = [
            {
                DOCUMENT_ID: node_with_score.node.node_id,
                DOCUMENT_SCORE: node_with_score.score,
                DOCUMENT_CONTENT: node_with_score.node.text,
                DOCUMENT_METADATA: node_with_score.node.metadata,
            }
            for node_with_score in payload[EventPayload.NODES]
        ]
    if EventPayload.PROMPT in payload:
        attributes[LLM_PROMPTS] = [payload[EventPayload.PROMPT]]
    if EventPayload.MESSAGES in payload:
        messages = payload[EventPayload.MESSAGES]
        # Messages is only relevant to the LLM invocation
        if event_type is CBEventType.LLM:
            attributes[LLM_INPUT_MESSAGES] = [
                _message_payload_to_attributes(message_data) for message_data in messages
            ]
        elif event_type is CBEventType.AGENT_STEP and len(messages):
            # the agent step contains a message that is actually the input
            # akin to the query_str
            attributes[INPUT_VALUE] = _message_payload_to_str(messages[0])
    if response := (payload.get(EventPayload.RESPONSE) or payload.get(EventPayload.COMPLETION)):
        attributes.update(_get_response_output(response))
        if raw := getattr(response, "raw", None):
            assert hasattr(raw, "get"), f"raw must be Mapping, found {type(raw)}"
            attributes.update(_get_output_messages(raw))
            if usage := raw.get("usage"):
                # OpenAI token counts are available on raw.usage but can also be
                # found in additional_kwargs. Thus the duplicate handling.
                attributes.update(_get_token_counts(usage))
        # Look for token counts in additional_kwargs of the completion payload
        # This is needed for non-OpenAI models
        if (additional_kwargs := getattr(response, "additional_kwargs", None)) is not None:
            attributes.update(_get_token_counts(additional_kwargs))
    if event_type is CBEventType.RERANKING:
        if EventPayload.TOP_K in payload:
            attributes[RERANKER_TOP_K] = payload[EventPayload.TOP_K]
        if EventPayload.MODEL_NAME in payload:
            attributes[RERANKER_MODEL_NAME] = payload[EventPayload.MODEL_NAME]
        if EventPayload.QUERY_STR in payload:
            attributes[RERANKER_QUERY] = payload[EventPayload.QUERY_STR]
        if nodes := payload.get(EventPayload.NODES):
            attributes[RERANKER_OUTPUT_DOCUMENTS if is_event_end else RERANKER_INPUT_DOCUMENTS] = [
                {
                    DOCUMENT_ID: node_with_score.node.node_id,
                    DOCUMENT_SCORE: node_with_score.score,
                    DOCUMENT_CONTENT: node_with_score.node.text,
                    DOCUMENT_METADATA: node_with_score.node.metadata,
                }
                for node_with_score in nodes
            ]
    if EventPayload.TOOL in payload:
        tool_metadata = cast(ToolMetadata, payload.get(EventPayload.TOOL))
        attributes[TOOL_NAME] = tool_metadata.name
        attributes[TOOL_DESCRIPTION] = tool_metadata.description
        attributes[TOOL_PARAMETERS] = tool_metadata.to_openai_tool()["function"]["parameters"]
    if EventPayload.SERIALIZED in payload:
        serialized = payload[EventPayload.SERIALIZED]
        if event_type is CBEventType.EMBEDDING:
            if model_name := serialized.get("model_name"):
                attributes[EMBEDDING_MODEL_NAME] = model_name
        if event_type is CBEventType.LLM:
            if model_name := serialized.get("model"):
                attributes[LLM_MODEL_NAME] = model_name
                attributes[LLM_INVOCATION_PARAMETERS] = json.dumps(
                    {
                        "model": model_name,
                        "temperature": serialized["temperature"],
                        "max_tokens": serialized["max_tokens"],
                        **serialized["additional_kwargs"],
                    }
                )
    return attributes


class OpenInferenceTraceCallbackHandler(BaseCallbackHandler):
    """Callback handler for storing LLM application trace data in OpenInference format.
    OpenInference is an open standard for capturing and storing AI model
    inferences. It enables production LLMapp servers to seamlessly integrate
    with LLM observability solutions such as Arize and Phoenix.

    For more information on the specification, see
    https://github.com/Arize-ai/open-inference-spec
    """

    def __init__(
        self,
        callback: Optional[Callable[[List[Span]], None]] = None,
        exporter: Optional[SpanExporter] = None,
    ) -> None:
        if (
            hasattr(llama_index, "__version__")
            and (version_triplet := extract_version_triplet(llama_index.__version__))
            and version_triplet < LLAMA_INDEX_MINIMUM_VERSION_TRIPLET
        ):
            raise NotImplementedError(
                f"minimum supported version of llama-index is "
                f"{'.'.join(map(str, LLAMA_INDEX_MINIMUM_VERSION_TRIPLET))}, "
                f"but yours is {llama_index.__version__}. "
                f"you can update to the latest using `pip install llama-index --upgrade`, "
                f"or install the minimum required for Phoenix using "
                f'`pip install "arize-phoenix[llama-index]"`',
            )
        super().__init__(event_starts_to_ignore=[], event_ends_to_ignore=[])
        self._tracer = Tracer(on_append=callback, exporter=exporter or HttpExporter())
        self._event_id_to_event_data: EventData = defaultdict(lambda: CBEventData())

    def _null_fallback(self, *args: Any, **kwargs: Any) -> None:
        return

    def _on_event_fallback(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: CBEventID = "",
        **kwargs: Any,
    ) -> CBEventID:
        return event_id or str(uuid4())

    @graceful_fallback(_on_event_fallback)
    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: CBEventID = "",
        parent_id: CBEventID = "",
        **kwargs: Any,
    ) -> CBEventID:
        event_id = event_id or str(uuid4())
        if parent_data := self._event_id_to_event_data.get(parent_id):
            trace_id = parent_data.trace_id
        else:
            trace_id = uuid4()
        event_data = self._event_id_to_event_data[event_id]
        event_data.name = event_type.value
        event_data.event_type = event_type
        event_data.parent_id = None if parent_id == "root" else parent_id
        event_data.span_id = event_id
        event_data.trace_id = trace_id
        event_data.start_event = CBEvent(
            event_type=event_type,
            payload=payload,
            id_=event_id,
        )
        # Parse the payload to extract the parameters
        if payload is not None:
            event_data.attributes.update(
                payload_to_semantic_attributes(event_type, payload),
            )

        return event_id

    @graceful_fallback(_null_fallback)
    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: CBEventID = "",
        **kwargs: Any,
    ) -> None:
        event_data = self._event_id_to_event_data[event_id]
        event_data.set_if_unset("name", event_type.value)
        event_data.set_if_unset("event_type", event_type)
        event_data.end_event = CBEvent(
            event_type=event_type,
            payload=payload,
            id_=event_id,
        )

        # Parse the payload to extract the parameters
        if payload is not None:
            event_data.attributes.update(
                payload_to_semantic_attributes(event_type, payload, is_event_end=True),
            )
            response = payload.get(EventPayload.RESPONSE)
            if _is_streaming_response(response):
                event_data.streaming_event = True
                response = _instrument_streaming_response(response, self._tracer, event_data)

    @graceful_fallback(_null_fallback)
    def start_trace(self, trace_id: Optional[str] = None) -> None:
        self._event_id_to_event_data = defaultdict(lambda: CBEventData())

    @graceful_fallback(_null_fallback)
    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[ChildEventIds] = None,
    ) -> None:
        if not trace_map:
            return  # TODO: investigate when empty or None trace_map is passed
        _add_spans_to_tracer(
            event_id_to_event_data=self._event_id_to_event_data,
            trace_map=trace_map,
            tracer=self._tracer,
        )
        self._event_id_to_event_data = defaultdict(lambda: CBEventData())

    def get_spans(self) -> Iterator[Span]:
        """
        Returns the spans stored in the tracer. This is useful if you are running
        LlamaIndex in a notebook environment and you want to inspect the spans.
        """
        return self._tracer.get_spans()


def _add_spans_to_tracer(
    event_id_to_event_data: EventData,
    trace_map: ChildEventIds,
    tracer: Tracer,
) -> None:
    """
    Adds event data to the tracer, where it is converted to a span and stored in a buffer.

    Args:
        event_id_to_event_data (EventData): A map of event IDs to event data.

        trace_map (ChildEventIds): A map of parent event IDs to child event IDs. The root event IDs
        are stored under the key "root".

        tracer (Tracer): The tracer that stores spans.
    """

    parent_child_id_stack: List[Tuple[Optional[SpanID], CBEventID]] = [
        (None, root_event_id) for root_event_id in trace_map["root"]
    ]
    while parent_child_id_stack:
        parent_span_id, event_id = parent_child_id_stack.pop()
        event_data = event_id_to_event_data[event_id]
        event_type = event_data.event_type
        attributes = event_data.attributes
        if not (start_event := event_data.start_event):
            # if the callback system has broken its contract by calling
            # on_event_end without on_event_start, do not create a span
            continue

        if event_type is CBEventType.LLM:
            while parent_child_id_stack:
                preceding_event_parent_span_id, preceding_event_id = parent_child_id_stack[-1]
                if preceding_event_parent_span_id != parent_span_id:
                    break
                preceding_event_data = event_id_to_event_data[preceding_event_id]
                if preceding_event_data.event_type is not CBEventType.TEMPLATING:
                    break
                parent_child_id_stack.pop()
                if preceding_event_start := preceding_event_data.start_event:
                    if preceding_payload := preceding_event_start.payload:
                        # Add template attributes to the LLM span to which they belong.
                        attributes.update(_template_attributes(preceding_payload))

        start_time = _timestamp_to_tz_aware_datetime(start_event.time)
        span_exceptions = _get_span_exceptions(event_data, start_time)
        if event_data.streaming_event:
            # Do not set the end time for streaming events so we can update the event later
            end_time = None
        else:
            end_time = _get_end_time(event_data, span_exceptions)
        start_time = start_time or end_time or datetime.now(timezone.utc)

        name = event_name if (event_name := event_data.name) is not None else "unknown"
        span_kind = _get_span_kind(event_type)
        span = tracer.create_span(
            name=name,
            span_kind=span_kind,
            trace_id=event_data.trace_id,
            start_time=start_time,
            end_time=end_time,
            status_code=SpanStatusCode.ERROR if span_exceptions else SpanStatusCode.OK,
            status_message="",
            parent_id=parent_span_id,
            attributes=attributes,
            events=sorted(span_exceptions, key=lambda event: event.timestamp) or None,
            conversation=None,
            span_id=UUID(event_data.span_id),
        )
        new_parent_span_id = span.context.span_id
        for new_child_event_id in trace_map.get(event_id, []):
            parent_child_id_stack.append((new_parent_span_id, new_child_event_id))


def _get_span_kind(event_type: Optional[CBEventType]) -> SpanKind:
    """Maps a CBEventType to a SpanKind.

    Args:
        event_type (CBEventType): LlamaIndex callback event type.

    Returns:
        SpanKind: The corresponding span kind.
    """
    if event_type is None:
        return SpanKind.UNKNOWN
    return {
        CBEventType.EMBEDDING: SpanKind.EMBEDDING,
        CBEventType.LLM: SpanKind.LLM,
        CBEventType.RETRIEVE: SpanKind.RETRIEVER,
        CBEventType.FUNCTION_CALL: SpanKind.TOOL,
        CBEventType.AGENT_STEP: SpanKind.AGENT,
        CBEventType.RERANKING: SpanKind.RERANKER,
    }.get(event_type, SpanKind.CHAIN)


def _message_payload_to_attributes(message: Any) -> Dict[str, Optional[str]]:
    if isinstance(message, ChatMessage):
        message_attributes = {
            MESSAGE_ROLE: message.role.value,
            MESSAGE_CONTENT: message.content,
        }
        # Parse the kwargs to extract the function name and parameters for function calling
        # NB: these additional kwargs exist both for 'agent' and 'function' roles
        if "name" in message.additional_kwargs:
            message_attributes[MESSAGE_NAME] = message.additional_kwargs["name"]
        if tool_calls := message.additional_kwargs.get("tool_calls"):
            assert isinstance(
                tool_calls, Iterable
            ), f"tool_calls must be Iterable, found {type(tool_calls)}"
            message_tool_calls = []
            for tool_call in tool_calls:
                if message_tool_call := dict(_get_tool_call(tool_call)):
                    message_tool_calls.append(message_tool_call)
            if message_tool_calls:
                message_attributes[MESSAGE_TOOL_CALLS] = message_tool_calls
        return message_attributes

    return {
        MESSAGE_ROLE: "user",  # assume user if not ChatMessage
        MESSAGE_CONTENT: str(message),
    }


def _message_payload_to_str(message: Any) -> Optional[str]:
    """Converts a message payload to a string, if possible"""
    if isinstance(message, ChatMessage):
        return message.content

    return str(message)


class _CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        try:
            return super().default(obj)
        except TypeError:
            if callable(as_dict := getattr(obj, "dict", None)):
                return as_dict()
            raise


def _get_response_output(response: Any) -> Iterator[Tuple[str, Any]]:
    """
    Gets output from response objects. This is needed since the string representation of some
    response objects includes extra information in addition to the content itself. In the
    case of an agent's ChatResponse the output may be a `function_call` object specifying
    the name of the function to call and the arguments to call it with.
    """
    if isinstance(response, ChatResponse):
        message = response.message
        if content := message.content:
            yield OUTPUT_VALUE, content
            yield OUTPUT_MIME_TYPE, MimeType.TEXT
        else:
            yield OUTPUT_VALUE, json.dumps(message.additional_kwargs, cls=_CustomJSONEncoder)
            yield OUTPUT_MIME_TYPE, MimeType.JSON
    elif isinstance(response, Response):
        yield OUTPUT_VALUE, response.response or ""
        yield OUTPUT_MIME_TYPE, MimeType.TEXT
    elif isinstance(response, StreamingResponse):
        # We cannot get the output from a streaming response without exhausting
        # the stream, so we initially return an empty string. Additional work is
        # needed to instrument the returned response object to update the span
        # with the actual response once the stream has been exhausted:
        # https://github.com/Arize-ai/phoenix/issues/1867
        yield OUTPUT_VALUE, ""
        yield OUTPUT_MIME_TYPE, MimeType.TEXT
    else:  # if the response has unknown type, make a best-effort attempt to get the output
        yield OUTPUT_VALUE, str(response)
        yield OUTPUT_MIME_TYPE, MimeType.TEXT


def _get_end_time(event_data: CBEventData, span_events: Iterable[SpanEvent]) -> Optional[datetime]:
    """
    A best-effort attempt to get the end time of an event.

    LlamaIndex's callback system does not guarantee that the on_event_end hook is always called, for
    example, when an error occurs mid-event.
    """
    if end_event := event_data.end_event:
        tz_naive_end_time = _timestamp_to_tz_naive_datetime(end_event.time)
    elif span_events:
        last_span_event = sorted(span_events, key=lambda event: event.timestamp)[-1]
        tz_naive_end_time = last_span_event.timestamp
    else:
        return None
    return _tz_naive_to_tz_aware_datetime(tz_naive_end_time)


def _get_span_exceptions(event_data: CBEventData, start_time: datetime) -> List[SpanException]:
    """Collects exceptions from the start and end events, if present."""
    span_exceptions = []
    for event in [event_data.start_event, event_data.end_event]:
        if event and (payload := event.payload) and (error := payload.get(EventPayload.EXCEPTION)):
            span_exceptions.append(
                SpanException(
                    message=str(error),
                    timestamp=start_time,
                    exception_type=type(error).__name__,
                    exception_stacktrace=get_stacktrace(error),
                )
            )
    return span_exceptions


def _timestamp_to_tz_aware_datetime(timestamp: str) -> datetime:
    """Converts a timestamp string to a timezone-aware datetime."""
    return _tz_naive_to_tz_aware_datetime(_timestamp_to_tz_naive_datetime(timestamp))


def _timestamp_to_tz_naive_datetime(timestamp: str) -> datetime:
    """Converts a timestamp string to a timezone-naive datetime."""
    return datetime.strptime(timestamp, TIMESTAMP_FORMAT)


def _tz_naive_to_tz_aware_datetime(timestamp: datetime) -> datetime:
    """Converts a timezone-naive datetime to a timezone-aware datetime."""
    return timestamp.replace(tzinfo=_LOCAL_TZINFO)


def _get_message(message: object) -> Iterator[Tuple[str, Any]]:
    if role := getattr(message, "role", None):
        assert isinstance(role, str), f"content must be str, found {type(role)}"
        yield MESSAGE_ROLE, role
    if content := getattr(message, "content", None):
        assert isinstance(content, str), f"content must be str, found {type(content)}"
        yield MESSAGE_CONTENT, content
    if tool_calls := getattr(message, "tool_calls", None):
        assert isinstance(
            tool_calls, Iterable
        ), f"tool_calls must be Iterable, found {type(tool_calls)}"
        message_tool_calls = []
        for tool_call in tool_calls:
            if message_tool_call := dict(_get_tool_call(tool_call)):
                message_tool_calls.append(message_tool_call)
        if message_tool_calls:
            yield MESSAGE_TOOL_CALLS, message_tool_calls


def _get_output_messages(raw: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
    assert hasattr(raw, "get"), f"raw must be Mapping, found {type(raw)}"
    if not (choices := raw.get("choices")):
        return
    assert isinstance(choices, Iterable), f"choices must be Iterable, found {type(choices)}"
    messages = [
        dict(_get_message(message))
        for choice in choices
        if (message := getattr(choice, "message", None)) is not None
    ]
    yield LLM_OUTPUT_MESSAGES, messages


def _get_token_counts(usage: Union[object, Mapping[str, Any]]) -> Iterator[Tuple[str, Any]]:
    """
    Yields token count attributes from a object or mapping
    """
    # Call the appropriate function based on the type of usage
    if isinstance(usage, Mapping):
        yield from _get_token_counts_from_mapping(usage)
    elif isinstance(usage, object):
        yield from _get_token_counts_from_object(usage)


def _get_token_counts_from_object(usage: object) -> Iterator[Tuple[str, Any]]:
    """
    Yields token count attributes from response.raw.usage
    """
    if (prompt_tokens := getattr(usage, "prompt_tokens", None)) is not None:
        yield LLM_TOKEN_COUNT_PROMPT, prompt_tokens
    if (completion_tokens := getattr(usage, "completion_tokens", None)) is not None:
        yield LLM_TOKEN_COUNT_COMPLETION, completion_tokens
    if (total_tokens := getattr(usage, "total_tokens", None)) is not None:
        yield LLM_TOKEN_COUNT_TOTAL, total_tokens


def _get_token_counts_from_mapping(
    usage_mapping: Mapping[str, Any],
) -> Iterator[Tuple[str, Any]]:
    """
    Yields token count attributes from a mapping (e.x. completion kwargs payload)
    """
    if (prompt_tokens := usage_mapping.get("prompt_tokens")) is not None:
        yield LLM_TOKEN_COUNT_PROMPT, prompt_tokens
    if (completion_tokens := usage_mapping.get("completion_tokens")) is not None:
        yield LLM_TOKEN_COUNT_COMPLETION, completion_tokens
    if (total_tokens := usage_mapping.get("total_tokens")) is not None:
        yield LLM_TOKEN_COUNT_TOTAL, total_tokens


def _template_attributes(payload: Dict[str, Any]) -> Iterator[Tuple[str, Any]]:
    """Yields template attributes if present"""
    if template := payload.get(EventPayload.TEMPLATE):
        yield LLM_PROMPT_TEMPLATE, template
    if template_vars := payload.get(EventPayload.TEMPLATE_VARS):
        yield LLM_PROMPT_TEMPLATE_VARIABLES, template_vars
        # TODO(maybe): other keys in the same payload
        # EventPayload.SYSTEM_PROMPT
        # EventPayload.QUERY_WRAPPER_PROMPT


def _get_tool_call(tool_call: object) -> Iterator[Tuple[str, Any]]:
    if function := getattr(tool_call, "function", None):
        if name := getattr(function, "name", None):
            assert isinstance(name, str), f"name must be str, found {type(name)}"
            yield TOOL_CALL_FUNCTION_NAME, name
        if arguments := getattr(function, "arguments", None):
            assert isinstance(arguments, str), f"arguments must be str, found {type(arguments)}"
            yield TOOL_CALL_FUNCTION_ARGUMENTS_JSON, arguments


def _is_streaming_response(response: Any) -> TypeGuard[StreamingResponse]:
    return isinstance(response, StreamingResponse)



import gzip
from typing import Protocol

from google.protobuf.message import DecodeError
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.status import HTTP_415_UNSUPPORTED_MEDIA_TYPE, HTTP_422_UNPROCESSABLE_ENTITY

import phoenix.trace.v1 as pb


class SupportsPutEvaluation(Protocol):
    def put(self, evaluation: pb.Evaluation) -> None:
        ...


class EvaluationHandler(HTTPEndpoint):
    queue: SupportsPutEvaluation

    async def post(self, request: Request) -> Response:
        content_type = request.headers.get("content-type")
        if content_type != "application/x-protobuf":
            return Response(
                content="Unsupported content type",
                status_code=HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            )
        body = await request.body()
        content_encoding = request.headers.get("content-encoding")
        if content_encoding == "gzip":
            body = gzip.decompress(body)
        elif content_encoding:
            return Response(
                content="Unsupported content encoding",
                status_code=HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            )
        evaluation = pb.Evaluation()
        try:
            evaluation.ParseFromString(body)
        except DecodeError:
            return Response(
                content="Request body is invalid",
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            )
        self.queue.put(evaluation)
        return Response()



import json
import uuid
from datetime import datetime
from typing import Any, Iterable, Iterator, List, Optional, cast

import pandas as pd
from pandas import DataFrame, read_parquet

from phoenix.datetime_utils import normalize_timestamps

from ..config import DATASET_DIR, GENERATED_DATASET_NAME_PREFIX
from .schemas import ATTRIBUTE_PREFIX, CONTEXT_PREFIX, Span
from .semantic_conventions import (
    DOCUMENT_METADATA,
    RERANKER_INPUT_DOCUMENTS,
    RERANKER_OUTPUT_DOCUMENTS,
    RETRIEVAL_DOCUMENTS,
)
from .span_evaluations import Evaluations, SpanEvaluations
from .span_json_decoder import json_to_span
from .span_json_encoder import span_to_json

# A set of columns that is required
REQUIRED_COLUMNS = [
    "name",
    "span_kind",
    "parent_id",
    "start_time",
    "end_time",
    "status_code",
    "status_message",
    "context.span_id",
    "context.trace_id",
]

RETRIEVAL_DOCUMENTS_COLUMN_NAME = f"{ATTRIBUTE_PREFIX}{RETRIEVAL_DOCUMENTS}"
RERANKER_INPUT_DOCUMENTS_COLUMN_NAME = f"{ATTRIBUTE_PREFIX}{RERANKER_INPUT_DOCUMENTS}"
RERANKER_OUTPUT_DOCUMENTS_COLUMN_NAME = f"{ATTRIBUTE_PREFIX}{RERANKER_OUTPUT_DOCUMENTS}"

DOCUMENT_COLUMNS = [
    RETRIEVAL_DOCUMENTS_COLUMN_NAME,
    RERANKER_INPUT_DOCUMENTS_COLUMN_NAME,
    RERANKER_OUTPUT_DOCUMENTS_COLUMN_NAME,
]


def normalize_dataframe(dataframe: DataFrame) -> "DataFrame":
    """Makes the dataframe have appropriate data types"""

    # Convert the start and end times to datetime
    dataframe["start_time"] = normalize_timestamps(dataframe["start_time"])
    dataframe["end_time"] = normalize_timestamps(dataframe["end_time"])
    return dataframe


def _delete_empty_document_metadata(documents: Any) -> Any:
    """
    Removes ambiguous and empty dicts from the documents list so the is object
    serializable to parquet
    """
    # If the documents is a list, iterate over them, check that the metadata is
    # a dict, see if it is empty, and if it's empty, delete the metadata
    if isinstance(documents, list):
        # Make a shallow copy of the keys
        documents = list(map(dict, documents))
        for document in documents:
            metadata = document.get(DOCUMENT_METADATA)
            if isinstance(metadata, dict) and not metadata:
                # Delete the metadata object since empty dicts are not serializable
                del document[DOCUMENT_METADATA]
    return documents


def get_serializable_spans_dataframe(dataframe: DataFrame) -> DataFrame:
    """
    Returns a dataframe that can be serialized to parquet. This means that
    the dataframe must not contain any unserializable objects. This function
    will delete any unserializable objects from the dataframe.
    """
    dataframe = dataframe.copy(deep=False)  # copy, don't mutate
    # Check if the dataframe has any document columns
    is_documents_column = dataframe.columns.isin(DOCUMENT_COLUMNS)
    for name, column in dataframe.loc[:, is_documents_column].items():  # type: ignore
        dataframe[name] = column.apply(_delete_empty_document_metadata)
    return dataframe


class TraceDataset:
    """
    A TraceDataset is a wrapper around a dataframe which is a flattened representation
    of Spans. The collection of spans trace the LLM application's execution.
    """

    name: str
    dataframe: pd.DataFrame
    evaluations: List[Evaluations] = []
    _data_file_name: str = "data.parquet"

    def __init__(
        self,
        dataframe: DataFrame,
        name: Optional[str] = None,
        evaluations: Iterable[Evaluations] = (),
    ):
        """
        Constructs a TraceDataset from a dataframe of spans. Optionally takes in
        evaluations for the spans in the dataset.

        Parameters
        __________
        dataframe: pandas.DataFrame
            the pandas dataframe containing the tracing data. Each row
            represents a span.
        evaluations: Optional[Iterable[SpanEvaluations]]
            an optional list of evaluations for the spans in the dataset. If
            provided, the evaluations can be materialized into a unified
            dataframe as annotations.
        """
        # Validate the the dataframe has required fields
        if missing_columns := set(REQUIRED_COLUMNS) - set(dataframe.columns):
            raise ValueError(
                f"The dataframe is missing some required columns: {', '.join(missing_columns)}"
            )
        self.dataframe = normalize_dataframe(dataframe)
        self.name = name or f"{GENERATED_DATASET_NAME_PREFIX}{str(uuid.uuid4())}"
        self.evaluations = list(evaluations)

    @classmethod
    def from_spans(cls, spans: List[Span]) -> "TraceDataset":
        """Creates a TraceDataset from a list of spans.

        Args:
            spans (List[Span]): A list of spans.

        Returns:
            TraceDataset: A TraceDataset containing the spans.
        """
        return cls(
            pd.json_normalize(
                (json.loads(span_to_json(span)) for span in spans),  # type: ignore
                max_level=1,
            )
        )

    def to_spans(self) -> Iterator[Span]:
        for _, row in self.dataframe.iterrows():
            is_attribute = row.index.str.startswith(ATTRIBUTE_PREFIX)
            attribute_keys = row.index[is_attribute]
            attributes = (
                row.loc[is_attribute]
                .rename(
                    {key: key[len(ATTRIBUTE_PREFIX) :] for key in attribute_keys},
                )
                .dropna()
                .to_dict()
            )
            is_context = row.index.str.startswith(CONTEXT_PREFIX)
            context_keys = row.index[is_context]
            context = (
                row.loc[is_context]
                .rename(
                    {key: key[len(CONTEXT_PREFIX) :] for key in context_keys},
                )
                .to_dict()
            )
            end_time: Optional[datetime] = cast(datetime, row.get("end_time"))
            if end_time is pd.NaT:
                end_time = None
            yield json_to_span(
                {
                    "name": row["name"],
                    "context": context,
                    "span_kind": row["span_kind"],
                    "parent_id": row.get("parent_id"),
                    "start_time": cast(datetime, row["start_time"]).isoformat(),
                    "end_time": end_time.isoformat() if end_time else None,
                    "status_code": row["status_code"],
                    "status_message": row.get("status_message") or "",
                    "attributes": attributes,
                    "events": row.get("events") or [],
                    "conversation": row.get("conversation"),
                }
            )

    @classmethod
    def from_name(cls, name: str) -> "TraceDataset":
        """Retrieves a dataset by name from the file system"""
        directory = DATASET_DIR / name
        df = read_parquet(directory / cls._data_file_name)
        return cls(df, name)

    def to_disc(self) -> None:
        """writes the data to disc"""
        directory = DATASET_DIR / self.name
        directory.mkdir(parents=True, exist_ok=True)
        get_serializable_spans_dataframe(self.dataframe).to_parquet(
            directory / self._data_file_name,
            allow_truncated_timestamps=True,
            coerce_timestamps="ms",
        )

    def append_evaluations(self, evaluations: Evaluations) -> None:
        """adds an evaluation to the traces"""
        # Append the evaluations to the list of evaluations
        self.evaluations.append(evaluations)

    def get_evals_dataframe(self) -> DataFrame:
        """
        Creates a flat dataframe of all the evaluations for the dataset.
        """
        return pd.concat(
            [
                evals.get_dataframe(prefix_columns_with_name=True)
                for evals in self.evaluations
                if isinstance(evals, SpanEvaluations)
            ],
            axis=1,
        )

    def get_spans_dataframe(self, include_evaluations: bool = True) -> DataFrame:
        """
        converts the dataset to a dataframe of spans. If evaluations are included,
        the evaluations are merged into the dataframe.

        Parameters
        __________
        include_evaluations: bool
            if True, the evaluations are merged into the dataframe
        """
        if not include_evaluations or not self.evaluations:
            return self.dataframe.copy()
        evals_df = self.get_evals_dataframe()
        # Make sure the index is set to the span_id
        df = self.dataframe.set_index("context.span_id", drop=False)
        return pd.concat([df, evals_df], axis=1)



from typing import Any, Optional

from .tracer import OpenInferenceTracer


class LangChainInstrumentor:
    """
    Instruments the OpenInferenceTracer for LangChain automatically by patching the
    BaseCallbackManager in LangChain.
    """

    def __init__(self, tracer: Optional[OpenInferenceTracer] = None) -> None:
        self._tracer = tracer if tracer is not None else OpenInferenceTracer()

    def instrument(self) -> None:
        try:
            from langchain.callbacks.base import BaseCallbackManager
        except ImportError:
            # Raise a cleaner error if LangChain is not installed
            raise ImportError(
                "LangChain is not installed. Please install LangChain first to use the instrumentor"
            )

        source_init = BaseCallbackManager.__init__

        # Keep track of the source init so we can tell if the patching occurred
        self._source_callback_manager_init = source_init

        tracer = self._tracer

        # Patch the init method of the BaseCallbackManager to add the tracer
        # to all callback managers
        def patched_init(self: BaseCallbackManager, *args: Any, **kwargs: Any) -> None:
            source_init(self, *args, **kwargs)
            self.add_handler(tracer, True)

        BaseCallbackManager.__init__ = patched_init
